// Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"
#include "LMBulletInfo.h"
#include "Camera/CameraComponent.h"
#include "GameFramework/Character.h"
#include "LMPlayer.generated.h"

class UCameraComponent;
class ULMWeaponManager;
class ULMHealthComponent;
class UMaterialInstanceDynamic;
class ALMAmmo;

DECLARE_DYNAMIC_MULTICAST_DELEGATE(FOnRequestUnpauseSignature);

UCLASS()
class LEGUMEMIX_API ALMPlayer : public ACharacter
{
	GENERATED_BODY()

public:
	// Sets default values for this character's properties
	ALMPlayer();
	
	UFUNCTION(BlueprintCallable)
	bool PickUpAmmo(ALMAmmo* Ammo);

	/**
	 * Gets the number of ammo from the given type.
	 * @param AmmoType The ammo type to get.
	 * @return The amount of ammo from the given type.
	 */
	UFUNCTION(BlueprintCallable)
	int GetAmmoCount(EAmmoType AmmoType) const;

	/**
	 * Removes a given amount of ammo from the type.
	 * @param AmmoType The type of ammo to remove.
	 * @param Count The amount of ammo to remove.
	 * @return The amount that was removed.
	 */
	UFUNCTION(BlueprintCallable)
	int RemoveAmmo(EAmmoType AmmoType, int Count) const;

protected:
	// Called when the game starts or when spawned
	virtual void BeginPlay() override;

	// Called every frame
	virtual void Tick(float DeltaTime) override;

public:	
	UFUNCTION(BlueprintCallable)
	ULMWeaponManager* GetWeaponManager() { return WeaponManager; }

	UFUNCTION(BlueprintCallable)
	void SetWeaponManager(ULMWeaponManager* Manager);

	UFUNCTION(BlueprintCallable)
	void PlayAnimation(UAnimMontage* Animation);

	UFUNCTION(BlueprintCallable)
	void FireBullets(const FLMBulletInfo Settings);

	UFUNCTION(BlueprintCallable)
	FVector GetWeaponFiringOrigin();

	UFUNCTION(BlueprintCallable)
	FVector GetAimVector();

	UFUNCTION(BlueprintCallable, BlueprintImplementableEvent)
	void WeaponFired();
	UFUNCTION(BlueprintCallable, BlueprintImplementableEvent)
	void WeaponReloaded();
	UFUNCTION(BlueprintCallable, BlueprintImplementableEvent)
	void WeaponSwitched();
	UFUNCTION(BlueprintCallable, BlueprintImplementableEvent)
	void OnAmmoPickup();

	UFUNCTION(BlueprintImplementableEvent)
	void OnEnnemyHit();
	
	UFUNCTION(BlueprintImplementableEvent, Category=Legumix)
	void OnFire();
	
	UFUNCTION(BlueprintImplementableEvent, Category=Legumix)
	void OnReload();

	UFUNCTION(BlueprintNativeEvent, BlueprintCallable)
	void OnPause(bool Paused);

public:
	UFUNCTION(BlueprintImplementableEvent)
	UCameraComponent* GetCamera();
	UCameraComponent* GetCam();

	UFUNCTION(BlueprintImplementableEvent)
	USkeletalMeshComponent* GetArmsBP();
	USkeletalMeshComponent* GetArms();
	
	UPROPERTY(BlueprintCallable, BlueprintAssignable, Category=Legumix)
	FOnRequestUnpauseSignature OnRequestedUnpause;

private:
	UPROPERTY(EditAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	TObjectPtr<ULMWeaponManager> WeaponManager;

	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	TObjectPtr<USkeletalMeshComponent> ArmsMesh;

	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	TObjectPtr<UCameraComponent> Camera;

	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	TObjectPtr<ULMHealthComponent> HealthComponent;

	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category=Legumix, meta = (AllowPrivateAccess = true))
	TObjectPtr<UMaterialInstanceDynamic> DamageDynamicMaterialInstance;

	/**
	 * Is a hit period in progress
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category=Legumix, meta = (AllowPrivateAccess = true))
	bool bHitPeriodInProgress;
	
	/**
	 * Length of the period in which the damage texture effect will be displayed on player screen.
	 */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category=Legumix, meta = (AllowPrivateAccess = true))
	double HitPeriodLengthInSeconds;
	
	/**
	 * Time of last damage taken.
	 * If the time elapsed since is superior to the HitPeriodLengthInSeconds variable, ends the Hit period
	 * and so the displaying the damage texture effect  on player screen
	 */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	double TimeOfTheLastDamage;

	/**
	 * Cumulated damages taken by the player on the current hit period
	 * To calculate view occlusion percent for the damage texture effect 
	 */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	float CumulatedDamagesOnCurrentHitPeriod;

	/**
	 * Player health at the current hit period beginning
	 * To calculate view occlusion percent for the damage texture effect 
	 */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	float HealthAtTheCurrentHitPeriodBeginning;

	/**
	 * Percent of the player screen occupied with damage texture effect to reach 
	 */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	float NextPlayerViewOcclusionPercent;

	/**
	 * Previous percent of the player screen occupied with damage texture effect
	 * mandatory to make interpolation between this core and the current one
	 */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category=Legumix, meta = (AllowPrivateAccess = true))
	float PlayerViewOcclusionPercent;
	
	/**
	 * Percent of the player screen occupied with damage texture effect 
	 */
	UPROPERTY(VisibleAnywhere, BlueprintReadWrite, Category=Legumix, meta = (AllowPrivateAccess = true))
	float AlphaLerpForPlayerViewOcclusionPercent;

	/**
	 * Actualize all parameters needed to calculate PlayerViewOcclusionPercent 
	 */
	UFUNCTION(BlueprintCallable, Category=PostProcess)
	void SetDisplayDamageParameters(float Health, float Damage);
		
	/**
	 * Determine Percent of the player screen occupied with damage texture effect 
	 */
	UFUNCTION(BlueprintCallable, Category=PostProcess)
	void SetPlayerViewOcclusionPercent();



private:
#if WITH_EDITORONLY_DATA
	/** If set, bullet debug will be drawn. */
	UPROPERTY(EditAnywhere, Category=Legumix, meta = (AllowPrivateAccess = true))
	bool bDrawBulletDebug = false;
#endif

private:
	FRandomStream SpreadStream;
};
